# -*- mode: tcl; indent-tabs-mode: nil -*-
# ----------------------------------------------------------------------
#  COMPONENT: geomapcamera - holds camera information for a geomap
#
# ======================================================================
#  AUTHOR:  Derrick Kearney, Purdue University
#  Copyright (c) 2004-2015  HUBzero Foundation, LLC
#
#  See the file "license.terms" for information on usage and
#  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# ======================================================================
package require Itcl
package require BLT

namespace eval Rappture {
    # forward declaration
}


itcl::class Rappture::GeoMapCamera {
    constructor {args} {
        # defined below
    }
    destructor {
        # defined below
    }

    private variable _viewpoints
    private variable _initial
    private variable _current

    public variable throw false
    public variable duration 1.0

    private method ViewpointCreate { name args }
    private method ViewpointCreateXYZ { name x y z args }
    private method ViewpointDelete { name }
    private method ViewpointNames { }
    private method ViewpointSave { name }
    private method ViewpointSettings { name }
    # private method ZoomIn { {factor -0.25} }
    # private method ZoomOut { {factor 0.25} }
    # private method ZoomReset {}
    # private method ZoomLayer { layerName }
    # private method ZoomDistance { distance }
    # private method ZoomExtent { xmin ymin xmax ymax } ;# duration srs
    private method GotoLocation { args } ;# using viewpoint args
    private method GotoViewpoint { viewpointName } ;# using viewpoint args
    private method GotoScreenCoords { x y {zoomFactor 1.0} }

    public method reset {}
    public method viewpoint { option args }

    # camera motion commands
    # public method zoom { option args }
    public method goto { args }
    # public method orient { args }
    # public method pan { x y }
    # public method rotate { x y }


    public method export { args }
}

# ----------------------------------------------------------------------
# CONSTRUCTOR
#
# USAGE:
# Rappture::GeoMapCamera #auto
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::constructor {args} {

    array set _viewpoints ""
    set _initial [namespace which [Rappture::GeoMapViewpoint #auto]]
    set _current [namespace which [Rappture::GeoMapViewpoint #auto $_initial]]
    eval configure $args
}


# ----------------------------------------------------------------------
# DESTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::destructor {} {
    # delete our viewpoint objects
    foreach viewpointName [ViewpointNames] {
        ViewpointDelete $viewpointName
    }
    # clear the viewpoints array
    unset _viewpoints

    # delete the initial viewpoint command
    itcl::delete object ${_initial}
    unset _initial

    # delete the current viewpoint command
    itcl::delete object ${_current}
    unset _current
}


# ----------------------------------------------------------------------
# ViewpointCreate: create and save a new viewpoint
#
# This method accepts all optional flags of the Rappture::GeoMapViewpoint
# class like -x, -y, -z, -heading, -pitch, -distance, -srs, -vertDatum,
# -label, and -description or another Rappture::GeoMapViewpoint object
# to be initialized from.
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::ViewpointCreate { name args } {

    set _viewpoints($name) [ \
        namespace which [ \
            eval Rappture::GeoMapViewpoint #auto $args]]

    return $name
}


# ----------------------------------------------------------------------
# ViewpointCreateXYZ: create and save a new viewpoint from x y z
#   camera coordinates
#
# This method also accepts optional flags of the Rappture::GeoMapViewpoint
# class like -heading, -pitch, -distance, -srs, -vertDatum, -label,
# and -description
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::ViewpointCreateXYZ { name x y z args } {

    # initialize the new viewpoint using _initial
    # this allows us to set similar defaults for optional flags like -heading
    set _viewpoints($name) [ \
        namespace which [ \
            Rappture::GeoMapViewpoint #auto ${_initial}]]

    # overwrite the _initial settings with user's settings.
    eval $_viewpoints($name) configure -x $x -y $y -z $z $args

    return $name
}


# ----------------------------------------------------------------------
# ViewpointDelete: remove a named viewpoint
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::ViewpointDelete { name } {
    if {![info exists _viewpoints($name)]} {
        error "bad value \"$name\": not a named viewpoint"
    }

    itcl::delete object $_viewpoints($name)
    unset _viewpoints($name)

    return $name
}


# ----------------------------------------------------------------------
# ViewpointNames: retrieve the names of all saved viewpoints
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::ViewpointNames { } {
    return [array names _viewpoints]
}


# ----------------------------------------------------------------------
# ViewpointSave: name and save the current viewpoint
#
# if the name already exists, it overwrites the stored viewpoint
# if a named viewpoint is already loaded, a second viewpoint is created
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::ViewpointSave { name } {

    if {[string equal $name "_initial_"] == 1} {
        # allow users to set/change the initial viewpoint
        ${_initial} import ${_current}
    } else {
        if {[info exists _viewpoints($name)]} {
            ViewpointDelete $name
        }
        ViewpointCreate $name ${_current}
    }

    return $name
}


# ----------------------------------------------------------------------
# ViewpointSettings: retrieve the settings for a named viewpoints
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::ViewpointSettings { name } {

    set settings ""
    set v ""

    if {[string equal $name "_initial_"] == 1} {
        set v $_initial
    } elseif {[string equal $name "_current_"] == 1} {
        set v $_current
    } elseif {[info exists _viewpoints($name)]} {
        set v $_viewpoints($name)
    } else {
        error "bad value \"$name\": not a named viewpoint"
    }

    set t [$v export]
    set settings [$t get root]

    return $settings
}


# ----------------------------------------------------------------------
# viewpoint: manage camera viewpoints
#
#   USAGE:
#       viewpoint create name x y z -heading <heading> -pitch <pitch> \
#           -distance <distance> -srs <srs> -vertdatum <vertdatum>
#       viewpoint delete name
#       viewpoint names
#       viewpoint save
#       viewpoint settings name
#
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::viewpoint {option args} {
    set result ""
    switch -- $option {
        "create" {
            set result [eval ViewpointCreateXYZ $args]
        }
        "delete" {
            set result [eval ViewpointDelete $args]
        }
        "names" {
            set result [ViewpointNames]
        }
        "save" {
            set result [eval ViewpointSave $args]
        }
        "settings" {
            set result [eval ViewpointSettings $args]
        }
        default {
            error "unknown layer option \"$option\""
        }
    }
    return $result
}


# ----------------------------------------------------------------------
# reset: set the camera properties to widget's default viewpoint
#
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::reset {} {
    ${_current} import ${_initial}
}


# ----------------------------------------------------------------------
# GotoLocation: expliticly set camera's current viewpoint.
#
#   Uses viewpoint flags.
#
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::GotoLocation { args } {
    eval ${_current} configure $args
}


# ----------------------------------------------------------------------
# GotoViewpoint: adjust the camera to a named viewpoint.
#
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::GotoViewpoint { name } {
    if {[info exists _viewpoints($name)] == 0} {
        set names [ViewpointNames]
        error "bad value \"$name\": should be a named viewpoint in $names"
    }
    ${_current} import $_viewpoints($name)
}


# ----------------------------------------------------------------------
# GotoScreenCoords: adjust the camera using screen coordinates.
#
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::GotoScreenCoords {x y {zoomFactor 1.0}} {
    if {[string is double $x] == 0} {
        error "bad value \"$x\": x should be a double"
    }
    if {[string is double $y] == 0} {
        error "bad value \"$y\": y should be a double"
    }
    if {[string is double $zoomFactor] == 0} {
        error "bad value \"$zoomFactor\": zoomFactor should be a double"
    }
    if {$zoomFactor < 0.0} {
        error "bad value \"$zoomFactor\": zoomFactor should be a positive double"
    }

# FIXME: hookup to "camera go <x> <y> <?zoomFactor?> <?duration?>"
}


# ----------------------------------------------------------------------
# goto
#
#   This public method is called by clients using this widget to manage
#   camera viewpoints.
#
#       goto location -x <x> -y <y> -z <z> \
#           -heading <heading> -pitch <pitch> \
#           -distance <distance> -srs <srs> -vertdatum <vertdatum>
#       goto viewpoint <name>
#       goto screencoords <x> <y> ?<zoom>?
#
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapCamera::goto {option args} {
    set result ""
    switch -- $option {
        "location" {
            set result [eval GotoLocation $args]
        }
        "viewpoint" {
            set result [eval GotoViewpoint $args]
        }
        "screencoords" {
            set result [eval GotoScreenCoords $args]
        }
        default {
            error "unknown layer option \"$option\""
        }
    }
    return $result
}


# ----------------------------------------------------------------------
# throw: enable / disable throw inertia for panning / rotating
# ----------------------------------------------------------------------
itcl::configbody Rappture::GeoMapCamera::throw {

    if {[string is boolean $throw] == 0} {
        error "bad value \"$throw\": throw should be a boolean"
    }
}


# ----------------------------------------------------------------------
# duration: get/set camera movement durations
# ----------------------------------------------------------------------
itcl::configbody Rappture::GeoMapCamera::duration {

    if {[string is double $duration] == 0} {
        error "bad value \"$duration\": duration should be a double"
    }
}


# ----------------------------------------------------------------------
# export: clients use this public method to serialize the object to a
#       blt::tree object
#
# export -format blt_tree
#
# ----------------------------------------------------------------------
itcl::body Rappture::GeoMapLayer::export {args} {
    set valids "-format"
    set format "blt_tree"

    while {[llength $args] > 0} {
        set flag [lindex $args 0]
        switch -- $flag {
            "-format" {
                if {[llength $args] > 1} {
                    set format [lindex $args 1]
                    set args [lrange $args 2 end]
                } else {
                    error "wrong number args: should be ?-format <format>?"
                }
            }
            default {
                error "invalid option \"$flag\": should be one of $valids"
            }
        }
    }

    set valids "blt_tree"

    switch -- $format {
        "blt_tree" {
            set t [blt::tree create]
#            $_provider exportToBltTree $t
#            $t set root \
#                label $label \
#                description $description \
#                attribution $attribution \
#                visible $visibility \
#                opacity $opacity \

        }
        default {
            error "bad format \"$format\": should be one of $valids"
        }
    }

    return $t
}

